@@ -6,9 +6,12 @@ import (
 	"time"
 
 	"github.com/influxdata/flux"
+	"github.com/influxdata/flux/codes"
 	"github.com/influxdata/flux/execute"
 	"github.com/influxdata/flux/execute/executetest"
+	"github.com/influxdata/flux/internal/errors"
 	"github.com/influxdata/flux/internal/gen"
+	"github.com/influxdata/flux/internal/operation"
 	"github.com/influxdata/flux/memory"
 	"github.com/influxdata/flux/querytest"
 	"github.com/influxdata/flux/stdlib/influxdata/influxdb"
@@ -20,12 +23,12 @@ func TestPivot_NewQuery(t *testing.T) {
 		{
 			Name: "pivot [_measurement, _field] around _time",
 			Raw:  `from(bucket:"testdb") |> range(start: -1h) |> pivot(rowKey: ["_time"], columnKey: ["_measurement", "_field"], valueColumn: "_value")`,
-			Want: &flux.Spec{
-				Operations: []*flux.Operation{
+			Want: &operation.Spec{
+				Operations: []*operation.Node{
 					{
 						ID: "from0",
 						Spec: &influxdb.FromOpSpec{
-							Bucket: "testdb",
+							Bucket: influxdb.NameOrID{Name: "testdb"},
 						},
 					},
 					{
@@ -52,7 +55,7 @@ func TestPivot_NewQuery(t *testing.T) {
 						},
 					},
 				},
-				Edges: []flux.Edge{
+				Edges: []operation.Edge{
 					{Parent: "from0", Child: "range1"},
 					{Parent: "range1", Child: "pivot2"},
 				},
@@ -73,33 +76,13 @@ func TestPivot_NewQuery(t *testing.T) {
 	}
 }
 
-func TestPivotOperation_Marshaling(t *testing.T) {
-	data := []byte(`{
-		"id":"pivot",
-		"kind":"pivot",
-		"spec":{
-			"rowKey":["_time"],
-			"columnKey":["_measurement", "_field"], 
-			"valueColumn":"_value"
-		}
-	}`)
-	op := &flux.Operation{
-		ID: "pivot",
-		Spec: &universe.PivotOpSpec{
-			RowKey:      []string{"_time"},
-			ColumnKey:   []string{"_measurement", "_field"},
-			ValueColumn: "_value",
-		},
-	}
-	querytest.OperationMarshalingTestHelper(t, data, op)
-}
-
 func TestPivot_Process(t *testing.T) {
 	testCases := []struct {
-		name string
-		spec *universe.PivotProcedureSpec
-		data []flux.Table
-		want []*executetest.Table
+		name    string
+		spec    *universe.PivotProcedureSpec
+		data    []flux.Table
+		want    []*executetest.Table
+		wantErr error
 	}{
 		{
 			name: "_field flatten case one table",
@@ -740,6 +723,62 @@ func TestPivot_Process(t *testing.T) {
 				},
 			},
 		},
+		{
+			name: "missing value column",
+			spec: &universe.PivotProcedureSpec{
+				RowKey:      []string{"_time"},
+				ColumnKey:   []string{"_field"},
+				ValueColumn: "_value",
+			},
+			data: []flux.Table{
+				&executetest.Table{
+					KeyCols: []string{"_measurement"},
+					ColMeta: []flux.ColMeta{
+						{Label: "_time", Type: flux.TTime},
+						{Label: "_new_value", Type: flux.TFloat},
+						{Label: "_measurement", Type: flux.TString},
+						{Label: "_field", Type: flux.TString},
+					},
+					Data: [][]interface{}{
+						{execute.Time(1), 1.0, "m1", "f1"},
+						{execute.Time(1), 2.0, "m1", "f2"},
+						{execute.Time(2), 3.0, "m1", "f1"},
+						{execute.Time(2), 4.0, "m1", "f2"},
+					},
+				},
+			},
+			wantErr: errors.New(codes.Invalid, "specified value column does not exist in table: _value"),
+		},
+		{
+			name: "column name conflict",
+			spec: &universe.PivotProcedureSpec{
+				RowKey:      []string{"_time"},
+				ColumnKey:   []string{"_field"},
+				ValueColumn: "_value",
+			},
+			data: []flux.Table{
+				&executetest.Table{
+					KeyCols: []string{"_measurement", "_field", "f1"},
+					ColMeta: []flux.ColMeta{
+						{Label: "_time", Type: flux.TTime},
+						{Label: "_value", Type: flux.TFloat},
+						{Label: "_measurement", Type: flux.TString},
+						{Label: "_field", Type: flux.TString},
+						{Label: "f1", Type: flux.TString},
+					},
+					Data: [][]interface{}{
+						{execute.Time(1), 1.0, "m1", "f1", "foo"},
+						{execute.Time(1), 2.0, "m1", "f2", "foo"},
+						{execute.Time(2), 3.0, "m1", "f1", "foo"},
+						{execute.Time(2), 4.0, "m1", "f2", "foo"},
+					},
+				},
+			},
+			wantErr: errors.New(
+				codes.Invalid,
+				`value "f1" appears in a column key column, but a column named "f1" already exists; consider renaming "f1" to something else before pivoting`,
+			),
+		},
 	}
 	for _, tc := range testCases {
 		tc := tc
@@ -748,7 +787,7 @@ func TestPivot_Process(t *testing.T) {
 				t,
 				tc.data,
 				tc.want,
-				nil,
+				tc.wantErr,
 				func(d execute.Dataset, c execute.TableBuilderCache) execute.Transformation {
 					return universe.NewPivotTransformation(d, c, tc.spec)
 				},
@@ -757,16 +796,162 @@ func TestPivot_Process(t *testing.T) {
 	}
 }
 
-func TestPivot2_Process(t *testing.T) {
+func TestSortedPivot_ProcessWithTags(t *testing.T) {
 	testCases := []struct {
 		name string
-		spec *universe.PivotProcedureSpec
-		data []flux.Table
+		spec *universe.SortedPivotProcedureSpec
+		data []flux.Table // test case data must be in groupKey sorted order
+		want []*executetest.Table
+	}{
+		{
+			name: "_field and tag with one measurement",
+			spec: &universe.SortedPivotProcedureSpec{
+				RowKey:      []string{"_time"},
+				ColumnKey:   []string{"_field"},
+				ValueColumn: "_value",
+			},
+			data: []flux.Table{
+				&executetest.Table{
+					KeyCols: []string{"_measurement", "_field", "host"},
+					ColMeta: []flux.ColMeta{
+						{Label: "_time", Type: flux.TTime},
+						{Label: "_value", Type: flux.TFloat},
+						{Label: "_measurement", Type: flux.TString},
+						{Label: "_field", Type: flux.TString},
+						{Label: "host", Type: flux.TString},
+					},
+					Data: [][]interface{}{
+						{execute.Time(1), 1.83, "system", "load1", "host.local"},
+						{execute.Time(2), 1.7, "system", "load1", "host.local"},
+						{execute.Time(3), 1.74, "system", "load1", "host.local"},
+						{execute.Time(4), 1.63, "system", "load1", "host.local"},
+						{execute.Time(5), 1.91, "system", "load1", "host.local"},
+						{execute.Time(6), 1.84, "system", "load1", "host.local"},
+					},
+				},
+				&executetest.Table{
+					KeyCols: []string{"_measurement", "_field", "host"},
+					ColMeta: []flux.ColMeta{
+						{Label: "_time", Type: flux.TTime},
+						{Label: "_value", Type: flux.TFloat},
+						{Label: "_measurement", Type: flux.TString},
+						{Label: "_field", Type: flux.TString},
+						{Label: "host", Type: flux.TString},
+					},
+					Data: [][]interface{}{
+						{execute.Time(1), 1.98, "system", "load15", "host.local"},
+						{execute.Time(2), 1.97, "system", "load15", "host.local"},
+						{execute.Time(3), 1.97, "system", "load15", "host.local"},
+						{execute.Time(4), 1.96, "system", "load15", "host.local"},
+						{execute.Time(5), 1.98, "system", "load15", "host.local"},
+						{execute.Time(6), 1.97, "system", "load15", "host.local"},
+					},
+				},
+				&executetest.Table{
+					KeyCols: []string{"_measurement", "_field", "host"},
+					ColMeta: []flux.ColMeta{
+						{Label: "_time", Type: flux.TTime},
+						{Label: "_value", Type: flux.TInt},
+						{Label: "_measurement", Type: flux.TString},
+						{Label: "_field", Type: flux.TString},
+						{Label: "host", Type: flux.TString},
+					},
+					Data: [][]interface{}{
+						{execute.Time(1), int64(95), "system", "load5", "host.local"},
+						{execute.Time(2), int64(92), "system", "load5", "host.local"},
+						{execute.Time(3), int64(92), "system", "load5", "host.local"},
+						{execute.Time(4), int64(89), "system", "load5", "host.local"},
+						{execute.Time(5), int64(94), "system", "load5", "host.local"},
+						{execute.Time(6), int64(93), "system", "load5", "host.local"},
+					},
+				},
+				&executetest.Table{
+					KeyCols: []string{"_measurement", "_field", "host"},
+					ColMeta: []flux.ColMeta{
+						{Label: "_time", Type: flux.TTime},
+						{Label: "_value", Type: flux.TInt},
+						{Label: "_measurement", Type: flux.TString},
+						{Label: "_field", Type: flux.TString},
+						{Label: "host", Type: flux.TString},
+					},
+					Data: [][]interface{}{
+						{execute.Time(1), int64(82), "swap", "used_p", "host.local"},
+						{execute.Time(2), int64(83), "swap", "used_p", "host.local"},
+						{execute.Time(3), int64(84), "swap", "used_p", "host.local"},
+						{execute.Time(4), int64(85), "swap", "used_p", "host.local"},
+						{execute.Time(5), int64(82), "swap", "used_p", "host.local"},
+					},
+				},
+			},
+			want: []*executetest.Table{
+				{
+					KeyCols: []string{"_measurement", "host"},
+					ColMeta: []flux.ColMeta{
+						{Label: "_time", Type: flux.TTime},
+						{Label: "_measurement", Type: flux.TString},
+						{Label: "host", Type: flux.TString},
+						{Label: "used_p", Type: flux.TInt},
+					},
+					Data: [][]interface{}{
+						{execute.Time(1), "swap", "host.local", int64(82)},
+						{execute.Time(2), "swap", "host.local", int64(83)},
+						{execute.Time(3), "swap", "host.local", int64(84)},
+						{execute.Time(4), "swap", "host.local", int64(85)},
+						{execute.Time(5), "swap", "host.local", int64(82)},
+					},
+				},
+				{
+					KeyCols: []string{"_measurement", "host"},
+					ColMeta: []flux.ColMeta{
+						{Label: "_time", Type: flux.TTime},
+						{Label: "_measurement", Type: flux.TString},
+						{Label: "host", Type: flux.TString},
+						{Label: "load1", Type: flux.TFloat},
+						{Label: "load15", Type: flux.TFloat},
+						{Label: "load5", Type: flux.TInt},
+					},
+					Data: [][]interface{}{
+						{execute.Time(1), "system", "host.local", 1.83, 1.98, int64(95)},
+						{execute.Time(2), "system", "host.local", 1.7, 1.97, int64(92)},
+						{execute.Time(3), "system", "host.local", 1.74, 1.97, int64(92)},
+						{execute.Time(4), "system", "host.local", 1.63, 1.96, int64(89)},
+						{execute.Time(5), "system", "host.local", 1.91, 1.98, int64(94)},
+						{execute.Time(6), "system", "host.local", 1.84, 1.97, int64(93)},
+					},
+				},
+			},
+		},
+	}
+	for _, tc := range testCases {
+		tc := tc
+		t.Run(tc.name, func(t *testing.T) {
+			executetest.ProcessTestHelper2(
+				t,
+				tc.data,
+				tc.want,
+				nil,
+				func(id execute.DatasetID, alloc memory.Allocator) (execute.Transformation, execute.Dataset) {
+					spec := *tc.spec
+					tr, d, err := universe.NewSortedPivotTransformation(context.Background(), spec, id, alloc)
+					if err != nil {
+						t.Fatal(err)
+					}
+					return tr, d
+				})
+		})
+	}
+}
+
+func TestSortedPivot_Process(t *testing.T) {
+	testCases := []struct {
+		name string
+		spec *universe.SortedPivotProcedureSpec
+		data []flux.Table // test case data must be in groupKey sorted order
 		want []*executetest.Table
 	}{
 		{
 			name: "_field flatten case one measurement",
-			spec: &universe.PivotProcedureSpec{
+			spec: &universe.SortedPivotProcedureSpec{
 				RowKey:      []string{"_time"},
 				ColumnKey:   []string{"_field"},
 				ValueColumn: "_value",
@@ -817,7 +1002,7 @@ func TestPivot2_Process(t *testing.T) {
 		},
 		{
 			name: "_field flatten case two measurements",
-			spec: &universe.PivotProcedureSpec{
+			spec: &universe.SortedPivotProcedureSpec{
 				RowKey:      []string{"_time"},
 				ColumnKey:   []string{"_field"},
 				ValueColumn: "_value",
@@ -907,7 +1092,7 @@ func TestPivot2_Process(t *testing.T) {
 		},
 		{
 			name: "_field flatten case two measurements different value type",
-			spec: &universe.PivotProcedureSpec{
+			spec: &universe.SortedPivotProcedureSpec{
 				RowKey:      []string{"_time"},
 				ColumnKey:   []string{"_field"},
 				ValueColumn: "_value",
@@ -997,7 +1182,7 @@ func TestPivot2_Process(t *testing.T) {
 		},
 		{
 			name: "dropping a column not in rowKey or groupKey",
-			spec: &universe.PivotProcedureSpec{
+			spec: &universe.SortedPivotProcedureSpec{
 				RowKey:      []string{"_time"},
 				ColumnKey:   []string{"_field"},
 				ValueColumn: "_value",
@@ -1050,7 +1235,7 @@ func TestPivot2_Process(t *testing.T) {
 		},
 		{
 			name: "group key doesn't change",
-			spec: &universe.PivotProcedureSpec{
+			spec: &universe.SortedPivotProcedureSpec{
 				RowKey:      []string{"_time"},
 				ColumnKey:   []string{"_field"},
 				ValueColumn: "_value",
@@ -1104,7 +1289,7 @@ func TestPivot2_Process(t *testing.T) {
 		},
 		{
 			name: "group key loses a member",
-			spec: &universe.PivotProcedureSpec{
+			spec: &universe.SortedPivotProcedureSpec{
 				RowKey:      []string{"_time"},
 				ColumnKey:   []string{"_field"},
 				ValueColumn: "_value",
@@ -1134,7 +1319,7 @@ func TestPivot2_Process(t *testing.T) {
 						{Label: "grouper", Type: flux.TString},
 					},
 					Data: [][]interface{}{
-						{execute.Time(1), 2.0, "m1", "f2", "B"},
+						{execute.Time(2), 4.0, "m1", "f2", "A"},
 					},
 				},
 				&executetest.Table{
@@ -1147,7 +1332,7 @@ func TestPivot2_Process(t *testing.T) {
 						{Label: "grouper", Type: flux.TString},
 					},
 					Data: [][]interface{}{
-						{execute.Time(2), 4.0, "m1", "f2", "A"},
+						{execute.Time(1), 2.0, "m1", "f2", "B"},
 					},
 				},
 			},
@@ -1182,7 +1367,7 @@ func TestPivot2_Process(t *testing.T) {
 		},
 		{
 			name: "group key loses all members. drops _value",
-			spec: &universe.PivotProcedureSpec{
+			spec: &universe.SortedPivotProcedureSpec{
 				RowKey:      []string{"_time"},
 				ColumnKey:   []string{"_field"},
 				ValueColumn: "grouper",
@@ -1247,7 +1432,7 @@ func TestPivot2_Process(t *testing.T) {
 		},
 		{
 			name: "_field flatten case one table with nulls",
-			spec: &universe.PivotProcedureSpec{
+			spec: &universe.SortedPivotProcedureSpec{
 				RowKey:      []string{"_time"},
 				ColumnKey:   []string{"_field"},
 				ValueColumn: "_value",
@@ -1260,11 +1445,12 @@ func TestPivot2_Process(t *testing.T) {
 						{Label: "_value", Type: flux.TFloat},
 						{Label: "_measurement", Type: flux.TString},
 						{Label: "_field", Type: flux.TString},
+						{Label: "gg", Type: flux.TString},
 					},
 					Data: [][]interface{}{
-						{execute.Time(1), 1.0, "m1", "f1"},
-						{execute.Time(2), 4.0, "m1", "f1"},
-						{execute.Time(3), nil, "m1", "f1"},
+						{execute.Time(1), 1.0, "m1", "f1", "gg1"},
+						{execute.Time(2), 4.0, "m1", "f1", "gg1"},
+						{execute.Time(3), nil, "m1", "f1", "gg1"},
 					},
 				},
 				&executetest.Table{
@@ -1314,6 +1500,16 @@ func TestPivot2_Process(t *testing.T) {
 				},
 			},
 		},
+		{
+			name: "pivot no data",
+			spec: &universe.SortedPivotProcedureSpec{
+				RowKey:      []string{"_time"},
+				ColumnKey:   []string{"_field"},
+				ValueColumn: "_value",
+			},
+			data: []flux.Table{},
+			want: nil,
+		},
 	}
 	for _, tc := range testCases {
 		tc := tc
@@ -1323,15 +1519,9 @@ func TestPivot2_Process(t *testing.T) {
 				tc.data,
 				tc.want,
 				nil,
-				func(id execute.DatasetID, alloc *memory.Allocator) (execute.Transformation, execute.Dataset) {
+				func(id execute.DatasetID, alloc memory.Allocator) (execute.Transformation, execute.Dataset) {
 					spec := *tc.spec
-					spec.IsKeyColumnFunc = func(label string) bool {
-						return true
-					}
-					spec.IsSortedByFunc = func(cols []string, desc bool) bool {
-						return true
-					}
-					tr, d, err := universe.NewPivotTransformation2(context.Background(), spec, id, alloc)
+					tr, d, err := universe.NewSortedPivotTransformation(context.Background(), spec, id, alloc)
 					if err != nil {
 						t.Fatal(err)
 					}
@@ -1341,21 +1531,15 @@ func TestPivot2_Process(t *testing.T) {
 	}
 }
 
-func TestPivot2_Process_VariousSchemas(t *testing.T) {
-	spec := universe.PivotProcedureSpec{
+func TestSortedPivot_Process_VariousSchemas(t *testing.T) {
+	spec := universe.SortedPivotProcedureSpec{
 		RowKey:      []string{"_time"},
 		ColumnKey:   []string{"_field"},
 		ValueColumn: "_value",
-		IsKeyColumnFunc: func(label string) bool {
-			return true
-		},
-		IsSortedByFunc: func(cols []string, desc bool) bool {
-			return true
-		},
 	}
-	mem := &memory.Allocator{}
+	mem := &memory.ResourceAllocator{}
 	id := executetest.RandomDatasetID()
-	tr, d, err := universe.NewPivotTransformation2(context.Background(), spec, id, mem)
+	tr, d, err := universe.NewSortedPivotTransformation(context.Background(), spec, id, mem)
 	if err != nil {
 		t.Fatal(err)
 	}
@@ -1363,7 +1547,7 @@ func TestPivot2_Process_VariousSchemas(t *testing.T) {
 	store := executetest.NewDataStore()
 	d.AddTransformation(store)
 
-	tables, err := gen.Input(gen.Schema{
+	tables, err := gen.Input(context.Background(), gen.Schema{
 		Tags: []gen.Tag{
 			{Name: "_measurement", Cardinality: 1},
 			{Name: "_field", Cardinality: 10},
@@ -1406,19 +1590,13 @@ func BenchmarkPivot(b *testing.B) {
 
 func benchmarkPivot(b *testing.B, n int) {
 	b.ReportAllocs()
-	spec := &universe.PivotProcedureSpec{
+	spec := &universe.SortedPivotProcedureSpec{
 		RowKey:      []string{execute.DefaultTimeColLabel},
 		ColumnKey:   []string{"_field"},
 		ValueColumn: execute.DefaultValueColLabel,
-		IsSortedByFunc: func(cols []string, desc bool) bool {
-			return true
-		},
-		IsKeyColumnFunc: func(label string) bool {
-			return true
-		},
 	}
 	executetest.ProcessBenchmarkHelper(b,
-		func(alloc *memory.Allocator) (flux.TableIterator, error) {
+		func(alloc memory.Allocator) (flux.TableIterator, error) {
 			schema := gen.Schema{
 				NumPoints: n,
 				Alloc:     alloc,
@@ -1429,13 +1607,13 @@ func benchmarkPivot(b *testing.B, n int) {
 					{Name: "t1", Cardinality: 50},
 				},
 			}
-			return gen.Input(schema)
+			return gen.Input(context.Background(), schema)
 		},
-		func(id execute.DatasetID, alloc *memory.Allocator) (execute.Transformation, execute.Dataset) {
+		func(id execute.DatasetID, alloc memory.Allocator) (execute.Transformation, execute.Dataset) {
 			// cache := execute.NewTableBuilderCache(alloc)
 			// d := execute.NewDataset(id, execute.DiscardingMode, cache)
 			// t := NewPivotTransformation(d, cache, spec)
-			t, d, err := universe.NewPivotTransformation2(context.Background(), *spec, id, alloc)
+			t, d, err := universe.NewSortedPivotTransformation(context.Background(), *spec, id, alloc)
 			if err != nil {
 				b.Fatal(err)
 			}
